#include "paddle/include/paddle_inference_api.h"

#include <chrono>
#include <iostream>
#include <memory>
#include <numeric>
#include <thread>

#include <gflags/gflags.h>
#include <glog/logging.h>

DEFINE_string(model_dir, "./mul_model", "Directory of the inference model.");
DEFINE_int32(warmup, 0, "warmup");
DEFINE_int32(repeats, 1, "repeats");
DEFINE_int32(thread_num, 1, "thread num");
DEFINE_int32(math_num, 1, "SetCpuMathLibraryNumThreads");
DEFINE_bool(profile, false, "native profile");
DEFINE_bool(mkldnn, false, "enable mkldnn");
DEFINE_int32(num, 1, "num");

namespace paddle {

using paddle::AnalysisConfig;

using Time = decltype(std::chrono::high_resolution_clock::now());
Time time() { return std::chrono::high_resolution_clock::now(); };
double time_diff(Time t1, Time t2) {
  typedef std::chrono::microseconds us;
  auto diff = t2 - t1;
  us counter = std::chrono::duration_cast<us>(diff);
  return counter.count() / 1000.0;
}

void PrepareConfig(AnalysisConfig *config) {
   config->SetModel(FLAGS_model_dir);
  //config->SetModel(FLAGS_model_dir + "/model", FLAGS_model_dir + "/params");
  config->DisableGpu();

  if (FLAGS_mkldnn) {
    config->EnableMKLDNN();
  }

  config->SetCpuMathLibraryNumThreads(FLAGS_math_num);

  // We use ZeroCopyTensor here, so we set config->SwitchUseFeedFetchOps(false)
  config->SwitchUseFeedFetchOps(false);

  if (FLAGS_profile) {
    config->EnableProfile();
  }
}

void work_thread(PaddlePredictor *predictor, int thread_id) {
  //auto predictor = predictor_master->Clone();

  int batch_size = 168;
  int channels = 200;
  int input_num = channels * batch_size;

  // prepare inputs
  float *input = new float[input_num];
  for (int i = 0; i < input_num; ++i) {
    input[i] = i % 255 * 0.8;
  }
  // memset(input, 0, input_num * sizeof(float));

  float* weight = new float[200 * 768];
  for (int i = 0; i < 200 * 768; ++i) {
    weight[i] = i % 10 * 0.1;
  }

  auto input_names = predictor->GetInputNames();
  auto input_t = predictor->GetInputTensor(input_names[0]);
  input_t->Reshape({batch_size, channels});
  input_t->copy_from_cpu(input);

  auto input_w = predictor->GetInputTensor(input_names[1]);
  input_w->Reshape({200, 768});
  input_w->copy_from_cpu(weight);

  for (size_t i = 0; i < FLAGS_warmup; ++i) {
    CHECK(predictor->ZeroCopyRun());
  }

  auto time1 = time();
  for (size_t i = 0; i < FLAGS_repeats; i++) {
    CHECK(predictor->ZeroCopyRun());
  }
  auto time2 = time();
  LOG(INFO) << "thread_id: " << thread_id << " batch: " << batch_size
            << " predict cost: "
            << time_diff(time1, time2) / static_cast<float>(FLAGS_repeats)
            << "ms" << std::endl;

  // get the output
  // std::vector<float> out_data;
  // auto output_names = predictor->GetOutputNames();
  // auto output_t = predictor->GetOutputTensor(output_names[0]);
  // std::vector<int> output_shape = output_t->shape();
  // int out_num = std::accumulate(output_shape.begin(), output_shape.end(), 1,
  //                               std::multiplies<int>());

  // out_data.resize(out_num);
  // output_t->copy_to_cpu(out_data.data());
  // for (size_t j = 0; j < out_num; ++j) {
  //   // LOG(INFO) << "output[" << j << "]: " << out_data[j];
  // }
  delete[] input;
  delete[] weight;
}
} // namespace paddle

int main(int argc, char **argv) {
  gflags::ParseCommandLineFlags(&argc, &argv, true);

  paddle::AnalysisConfig config;
  PrepareConfig(&config);
  auto main_predictor = paddle::CreatePaddlePredictor(config);
  std::vector<std::unique_ptr<paddle::PaddlePredictor>> predictor_vector;
  for (int i = 0; i < FLAGS_thread_num; i++) {
      predictor_vector.emplace_back(main_predictor->Clone());
  }

  std::vector<std::thread> threads;

  int num = FLAGS_num;
  while (num--) {
    for (int i = 0; i < FLAGS_thread_num; i++) {
      threads.emplace_back(std::thread(paddle::work_thread, predictor_vector[i].get(), i));
    }
    for (auto &t : threads) {
      t.join();
    }
    threads.clear();
  }
  return 0;
}
